'''Cobra protocol.

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio
import base64
import itertools
import json
import logging
import platform
from typing import Dict
from urllib.parse import parse_qs, urlparse

import ujson
import websockets

from cobras.common.apps_config import generateNonce
from cobras.common.auth_hash import computeHash
from cobras.common.cobra_types import JsonDict
from cobras.common.task_cleanup import addTaskCleanup
from cobras.common.throttle import Throttle
from cobras.common.version import getVersion
from cobras.server.connection_state import ConnectionState
from cobras.server.redis_connections import RedisConnections
from cobras.server.redis_kv_store import kvStoreRead
from cobras.server.redis_subscriber import (RedisSubscriberMessageHandlerClass,
                                            redisSubscriber, validatePosition)
from cobras.server.stream_sql import InvalidStreamSQLError, StreamSqlFilter


async def respond(state: ConnectionState, ws, app: Dict, data: JsonDict):
    response = json.dumps(data)
    logging.info(f"> {response}")

    try:
        await ws.send(response)
    except websockets.exceptions.ConnectionClosed as e:
        logging.warning(f'Trying to write in a closed connection: {e}')


async def handleAuth(state: ConnectionState, ws, app: Dict, pdu: JsonDict,
                     serializedPdu: bytes):
    try:
        secret = app['apps_config'].getRoleSecret(state.appkey, state.role)
    except KeyError:
        reason = 'invalid_role'
        success = False
    else:
        serverHash = computeHash(secret, state.nonce.encode('ascii'))
        clientHash = pdu.get('body', {}).get('credentials', {}).get('hash')

        state.log(f'server hash {serverHash}')
        state.log(f'client hash {clientHash}')

        success = (clientHash == serverHash)
        if not success:
            reason = 'challenge_failed'

    if success:
        response = {
            "action": "auth/authenticate/ok",
            "id": pdu.get('id', 1),
            "body": {},
        }

        state.authenticated = True
        state.permissions = app['apps_config'].getPermissions(
            state.appkey, state.role)
    else:
        logging.warning(f'auth error: {reason}')
        response = {
            "action": "auth/authenticate/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": "authentication_failed",
                "reason": reason
            }
        }
    await respond(state, ws, app, response)


async def badFormat(state: ConnectionState, ws, app: Dict, reason: str):
    response = {"body": {"error": "bad_schema", "reason": reason}}
    state.ok = False
    state.error = response
    await respond(state, ws, app, response)


def parseAppKey(path):
    '''
    Parse url
    path = /v2?appkey=FFFFFFFFFFFFEEEEEEEEEEEEE
    '''
    parseResult = urlparse(path)
    args = parse_qs(parseResult.query)
    appkey = args.get('appkey')
    if appkey is None or not isinstance(appkey, list) or len(appkey) != 1:
        return None

    appkey = appkey[0]
    return appkey


async def handleHandshake(state: ConnectionState, ws, app: Dict, pdu: JsonDict,
                          serializedPdu: bytes):
    authMethod = pdu.get('body', {}).get('method')
    if authMethod != 'role_secret':
        errMsg = f'invalid auth method: {authMethod}'
        logging.warning(errMsg)
        response = {
            "action": f"auth/handshake/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": errMsg
            }
        }
        await respond(state, ws, app, response)
        return

    role = pdu.get('body', {}).get('data', {}).get('role')
    state.role = role
    state.nonce = generateNonce()

    response = {
        "action": "auth/handshake/ok",
        "id": pdu.get('id', 1),
        "body": {
            "data": {
                "nonce": state.nonce,
                "version": getVersion(),
                "connection_id": state.connection_id,
                "node": platform.uname().node
            }
        }
    }
    await respond(state, ws, app, response)


async def handlePublish(state: ConnectionState, ws, app: Dict, pdu: JsonDict,
                        serializedPdu: bytes):
    '''Here we don't write back a result to the client for efficiency.
    Client doesn't really needs it.
    '''
    # Use plugins to transform input data, if any
    plugins = app.get('plugins')
    if plugins is not None:
        pdu = plugins.updateMsg(state.appkey, pdu)

    # Missing message
    message = pdu.get('body', {}).get('message')
    if message is None:
        errMsg = 'publish: empty message'
        logging.warning(errMsg)
        response = {
            "action": "rtm/publish/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": errMsg
            }
        }
        await respond(state, ws, app, response)
        return

    # Missing channels
    channel = pdu.get('body', {}).get('channel')
    channels = pdu.get('body', {}).get('channels')
    if channel is None and channels is None:
        errMsg = 'publish: no channel or channels field'
        logging.warning(errMsg)
        response = {
            "action": "rtm/publish/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": errMsg
            }
        }
        await respond(state, ws, app, response)
        return

    if channels is None:
        channels = [channel]

    batchPublish = app['apps_config'].isBatchPublishEnabled(state.appkey)

    # We could have metrics per channel
    for chan in channels:
        appkey = state.appkey

        try:
            pipelinedPublisher = \
                await app['pipelined_publishers'].get(appkey, chan)

            await pipelinedPublisher.push((appkey, chan, serializedPdu),
                                          batchPublish)
        except Exception as e:
            errMsg = f'publish: cannot connect to redis {e}'
            logging.warning(errMsg)
            response = {
                "action": "rtm/publish/error",
                "id": pdu.get('id', 1),
                "body": {
                    "error": errMsg
                }
            }
            await respond(state, ws, app, response)
            return

    response = {
        "action": "rtm/publish/ok",
        "id": pdu.get('id', 1),
        "body": {}
    }
    await respond(state, ws, app, response)

    # Stats
    app['stats'].updatePublished(state.role, len(serializedPdu))


async def handleSubscribe(state: ConnectionState, ws, app: Dict, pdu: JsonDict,
                          serializedPdu: bytes):
    '''
    Client doesn't really needs it.
    '''
    body = pdu.get('body', {})
    channel = body.get('channel')

    subscriptionId = body.get('subscription_id')

    if channel is None and subscriptionId is None:
        errMsg = 'missing channel and subscription_id'
        logging.warning(errMsg)
        response = {
            "action": "rtm/subscribe/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": errMsg
            }
        }
        await respond(state, ws, app, response)
        return

    maxSubs = app['max_subscriptions']
    if maxSubs >= 0 and len(state.subscriptions) + 1 > maxSubs:
        errMsg = f'subscriptions count over max limit: {maxSubs}'
        logging.warning(errMsg)
        response = {
            "action": "rtm/subscribe/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": errMsg
            }
        }
        state.ok = False
        state.error = response
        await respond(state, ws, app, response)
        return

    if channel is None:
        channel = subscriptionId

    if subscriptionId is None:
        subscriptionId = channel

    filterStr = body.get('filter')
    hasFilter = filterStr not in ('', None)

    try:
        streamSQLFilter = StreamSqlFilter(filterStr) if hasFilter else None
    except InvalidStreamSQLError:
        errMsg = f'Invalid SQL expression {filterStr}'
        logging.warning(errMsg)
        response = {
            "action": "rtm/subscribe/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": errMsg
            }
        }
        state.ok = False
        state.error = response
        await respond(state, ws, app, response)
        return

    if hasFilter:
        channel = streamSQLFilter.channel

    position = body.get('position')
    if not validatePosition(position):
        errMsg = f'Invalid position: {position}'
        logging.warning(errMsg)
        response = {
            "action": "rtm/subscribe/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": errMsg
            }
        }
        state.ok = False
        state.error = response
        await respond(state, ws, app, response)
        return

    response = {
        "action": "rtm/subscribe/ok",
        "id": pdu.get('id', 1),
        "body": {
            "position": "1519190184:559034812775",
            "subscription_id": subscriptionId
        },
    }

    class MessageHandlerClass(RedisSubscriberMessageHandlerClass):
        def __init__(self, args):
            self.cnt = 0
            self.cntPerSec = 0
            self.throttle = Throttle(seconds=1)
            self.ws = args['ws']
            self.subscriptionId = args['subscription_id']
            self.hasFilter = args['has_filter']
            self.streamSQLFilter = args['stream_sql_filter']
            self.appkey = args['appkey']
            self.serverStats = args['stats']
            self.state = args['state']
            self.subscribeResponse = args['subscribe_response']
            self.app = args['app']
            self.idIterator = itertools.count()

        def log(self, msg):
            self.state.log(msg)

        async def on_init(self, redisConnection):
            response = self.subscribeResponse
            if redisConnection is None:
                response['body']['action'] = 'rtm/subscribe/error'
            else:
                response['body']['redis_node'] = redisConnection.host
            await respond(self.state, self.ws, self.app, response)

        async def handleMsg(self, msg: dict, position: str,
                            payloadSize: int) -> bool:

            # Input msg is the full serialized publish pdu.
            # Extract the real message out of it.
            msg = msg.get('body', {}).get('message')

            self.serverStats.updateSubscribed(self.state.role, payloadSize)

            if self.hasFilter:
                filterOutput = self.streamSQLFilter.match(
                    msg.get('messages') or msg)  # noqa
                if not filterOutput:
                    return True
                else:
                    msg = filterOutput

            pdu = {
                "action": "rtm/subscription/data",
                "id": next(self.idIterator),
                "body": {
                    "subscription_id": self.subscriptionId,
                    "messages": [msg],
                    "position": position
                },
            }
            self.state.log(f"> {json.dumps(pdu)} at position {position}")

            await self.ws.send(json.dumps(pdu))

            self.cnt += 1
            self.cntPerSec += 1

            if self.throttle.exceedRate():
                return True

            self.state.log(f"#messages {self.cnt} msg/s {self.cntPerSec}")
            self.cntPerSec = 0

            return True

    appChannel = '{}::{}'.format(state.appkey, channel)

    redisConnections = RedisConnections(app['redis_urls'],
                                        app['redis_password'])

    task = asyncio.create_task(
        redisSubscriber(
            redisConnections, appChannel, position, MessageHandlerClass, {
                'ws': ws,
                'subscription_id': subscriptionId,
                'has_filter': hasFilter,
                'stream_sql_filter': streamSQLFilter,
                'appkey': state.appkey,
                'stats': app['stats'],
                'state': state,
                'subscribe_response': response,
                'app': app
            }))
    addTaskCleanup(task)

    key = subscriptionId + state.connection_id
    state.subscriptions[key] = (task, state.role)

    app['stats'].incrSubscriptions(state.role)


async def handleUnSubscribe(state: ConnectionState, ws, app: Dict,
                            pdu: JsonDict, serializedPdu: bytes):
    '''
    Cancel a subscription
    '''
    body = pdu.get('body', {})

    subscriptionId = body.get('subscription_id')
    if subscriptionId is None:
        errMsg = f'Body Missing subscriptionId'
        logging.warning(errMsg)
        response = {
            "action": "rtm/unsubscribe/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": errMsg
            }
        }
        await respond(state, ws, app, response)
        return

    key = subscriptionId + state.connection_id
    item = state.subscriptions.get(key, (None, None))
    task, _ = item
    if task is None:
        errMsg = f'Invalid subscriptionId: {subscriptionId}'
        logging.warning(errMsg)
        response = {
            "action": "rtm/unsubscribe/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": errMsg
            }
        }
        await respond(state, ws, app, response)
        return

    # Correct path
    response = {"action": "rtm/unsubscribe/ok", "id": pdu.get('id', 1)}
    await respond(state, ws, app, response)

    task.cancel()


async def closeAll(state: ConnectionState, app: Dict, params: JsonDict):
    websocketLists = []
    for connectionId, (state, websocket) in app['connections'].items():
        if connectionId != state.connection_id:
            websocketLists.append(websocket)

    for ws in websocketLists:
        await ws.close()  # should this be shielded ?

    return {'status': 'ok'}


async def toggleFileLogging(state: ConnectionState, app: Dict,
                            params: JsonDict):
    found = False
    for connectionId, (st, websocket) in app['connections'].items():
        if connectionId == params.get('connection_id', ''):
            st.fileLogging = not st.fileLogging
            found = True

    return {'found': found, 'params': params}


# FIXME: make sure that 'write' commands are coming from 'localhost'
async def handleAdminRpc(state: ConnectionState, ws, app: Dict, pdu: JsonDict,
                         serializedPdu: bytes):
    body = pdu.get('body', {})
    method = body.get('method')
    params = body.get('params', {})

    methods = {'close_all': closeAll, 'toggle_file_logging': toggleFileLogging}

    if method is None or method not in methods:
        errMsg = f'Invalid method'
        logging.warning(errMsg)
        response = {
            "action": "rpc/admin/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": errMsg
            }
        }
        await respond(state, ws, app, response)
        return

    func = methods.get(method)  # lookup callback
    response = await func(state, app, params)  # execute it

    responseBody = {"data": response}

    response = {
        "action": "rpc/admin/ok",
        "id": pdu.get('id', 1),
        "body": responseBody
    }
    await respond(state, ws, app, response)


# FIXME error handling
async def handleRead(state: ConnectionState, ws, app: Dict, pdu: JsonDict,
                     serializedPdu: bytes):

    body = pdu.get('body', {})
    position = body.get('position')
    channel = body.get('channel')

    appChannel = '{}::{}'.format(state.appkey, channel)

    redisConnections = RedisConnections(app['redis_urls'],
                                        app['redis_password'])

    message = await kvStoreRead(redisConnections, appChannel, position,
                                state.log)

    # Correct path
    response = {
        "action": "rtm/read/ok",
        "id": pdu.get('id', 1),
        "body": {
            "message": message
        }
    }
    await respond(state, ws, app, response)


async def handleWrite(state: ConnectionState, ws, app: Dict, pdu: JsonDict,
                      serializedPdu: bytes):
    # Missing message
    message = pdu.get('body', {}).get('message')
    if message is None:
        errMsg = 'write: empty message'
        logging.warning(errMsg)
        response = {
            "action": "rtm/write/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": errMsg
            }
        }
        await respond(state, ws, app, response)
        return

    # Missing channel
    channel = pdu.get('body', {}).get('channel')
    if channel is None:
        errMsg = 'write: missing channel field'
        logging.warning(errMsg)
        response = {
            "action": "rtm/write/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": errMsg
            }
        }
        await respond(state, ws, app, response)
        return

    # Extract the message. This is what will be published
    message = pdu['body']['message']

    appkey = state.appkey
    pipelinedPublisher = \
        await app['pipelined_publishers'].get(appkey, channel)

    await pipelinedPublisher.publishNow((appkey, channel, json.dumps(message)))

    # Stats
    app['stats'].updateWrites(state.role, len(serializedPdu))

    response = {
        "action": f"rtm/write/ok",
        "id": pdu.get('id', 1),
        "body": {}
    }
    await respond(state, ws, app, response)


def validatePermissions(permissions, action):
    group, sep, verb = action.partition('/')
    if group != 'rtm' or sep != '/':
        return True

    if verb == 'unsubscribe':
        return True
    
    return verb in permissions


AUTH_PREFIX = 'auth'
ACTION_HANDLERS_LUT = {
    f'{AUTH_PREFIX}/handshake': handleHandshake,
    f'{AUTH_PREFIX}/authenticate': handleAuth,
    'rtm/publish': handlePublish,
    'rtm/subscribe': handleSubscribe,
    'rtm/unsubscribe': handleUnSubscribe,
    'rtm/read': handleRead,
    'rtm/write': handleWrite,
    'rpc/admin': handleAdminRpc
}


async def processCobraMessage(state: ConnectionState, ws, app: Dict, msg: bytes):
    try:
        pdu: JsonDict = ujson.loads(msg)
    except ValueError:
        msgEncoded = base64.b64encode(msg)
        errMsg = f'malformed json pdu: base64: {msgEncoded} raw: {msg}'
        await badFormat(state, ws, app, errMsg)
        return

    state.log(f"< {msg}")

    action = pdu.get('action')
    if action is None:
        await badFormat(state, ws, app, f'missing action')
        return

    handler = ACTION_HANDLERS_LUT.get(action)
    if handler is None:
        await badFormat(state, ws, app, f'invalid action: {action}')
        return

    # Make sure the user is authenticated
    if not state.authenticated and not action.startswith(AUTH_PREFIX):
        errMsg = f'action "{action}" needs authentication'
        logging.warning(errMsg)
        response = {
            "action": f"{action}/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": errMsg
            }
        }
        await respond(state, ws, app, response)
        return

    # Make sure the user has permission to access given endpoint
    if not validatePermissions(state.permissions, action):
        errMsg = f'action "{action}": permission denied'
        logging.warning(errMsg)
        response = {
            "action": f"{action}/error",
            "id": pdu.get('id', 1),
            "body": {
                "error": errMsg
            }
        }
        await respond(state, ws, app, response)
        return

    # proceed with handling action
    await handler(state, ws, app, pdu, msg)



'''Health check.

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.

Subscribe to a channel, publish and make sure a message is received
'''
import sys
import urllib.request

import click
import uvloop

from cobras.client.credentials import (getDefaultRoleForApp,
                                       getDefaultSecretForApp)
from cobras.client.health_check import (getDefaultHealthCheckChannel,
                                        getDefaultHealthCheckHttpUrl,
                                        getDefaultHealthCheckUrl, healthCheck)
from cobras.common.superuser import preventRootUsage


@click.option('--url', default=getDefaultHealthCheckUrl())
@click.option('--http_url', default=getDefaultHealthCheckHttpUrl)
@click.option('--http', is_flag=True)
@click.option('--role', default=getDefaultRoleForApp('health'))
@click.option('--secret', default=getDefaultSecretForApp('health'))
@click.option('--channel', default=getDefaultHealthCheckChannel())
@click.option('--retry', is_flag=True)
@click.command()
def health(url, http_url, http, role, secret, channel, retry):
    '''Health check

    \b
    cobra health --http
    \b
    cobra health --http --http_url 'http://127.0.0.1:8765/health/'
    \b
    '''
    preventRootUsage()
    uvloop.install()

    if http:
        print('url:', http_url)
        with urllib.request.urlopen(http_url) as response:
            html = response.read()
            print(html.decode('utf8'), end='')
    else:
        try:
            healthCheck(url, role, secret, channel, retry)
        except ValueError as e:
            click.secho(f'System is unhealthy !!: {e}', fg='red')
            sys.exit(1)

    click.secho('System is healthy', fg='green')



'''Redis subscriber

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import logging
import asyncio
import re
import traceback
from abc import ABC, abstractmethod
from typing import Optional

import ujson

from cobras.server.redis_connections import RedisConnections

POSITION_PATTERN = re.compile('^(?P<id1>[0-9]+)-(?P<id2>[0-9]+)')


def validatePosition(position):
    if position is None or position == '$':
        return True

    return POSITION_PATTERN.match(position)


class RedisSubscriberMessageHandlerClass(ABC):
    def __init__(self, args):
        pass  # pragma: no cover

    @abstractmethod
    def log(self, msg):
        pass  # pragma: no cover

    @abstractmethod
    async def on_init(self):
        pass  # pragma: no cover

    @abstractmethod
    async def handleMsg(self, msg: dict, position: str,
                        payloadSize: int) -> bool:
        return True  # pragma: no cover


async def redisSubscriber(
        redisConnections: RedisConnections,
        pattern: str,
        position: Optional[str],
        messageHandlerClass: RedisSubscriberMessageHandlerClass,  # noqa
        obj):
    messageHandler = messageHandlerClass(obj)

    try:
        # Create connection
        connection = await redisConnections.create(pattern)
    except Exception as e:
        logging.error(f"subcriber: cannot connect to redis {e}")
        connection = None

    await messageHandler.on_init(connection)

    if connection is None:
        return messageHandler

    # lastId = '0-0'
    lastId = '$' if position is None else position

    try:
        # wait for incoming events.
        while True:
            results = await connection.xread([pattern],
                                             timeout=0,
                                             latest_ids=[lastId])

            for result in results:
                lastId = result[1]
                msg = result[2]
                data = msg[b'json']

                payloadSize = len(data)
                msg = ujson.loads(data)
                ret = await messageHandler.handleMsg(msg, lastId.decode(),
                                                     payloadSize)
                if not ret:
                    break

    except asyncio.CancelledError:
        messageHandler.log('Cancelling redis subscription')
        raise

    except Exception as e:
        messageHandler.log(e)
        messageHandler.log('Generic Exception caught in {}'.format(
            traceback.format_exc()))

    finally:
        messageHandler.log('Closing redis subscription')

        # When finished, close the connection.
        connection.close()

        return messageHandler


def runSubscriber(redisConnections: RedisConnections,
                  channel: str,
                  position: str,
                  messageHandlerClass,
                  obj=None):
    asyncio.get_event_loop().run_until_complete(
        redisSubscriber(redisConnections, channel, position,
                        messageHandlerClass, obj))



'''Monitor cobra

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import click
import uvloop

from cobras.client.credentials import (createCredentials, getDefaultRoleForApp,
                                       getDefaultSecretForApp)
from cobras.client.monitor import runMonitor
from cobras.common.apps_config import STATS_APPKEY
from cobras.common.superuser import preventRootUsage

DEFAULT_URL = f'ws://127.0.0.1:8765/v2?appkey={STATS_APPKEY}'


@click.command()
@click.option('--url', default=DEFAULT_URL)
@click.option('--role', default=getDefaultRoleForApp('stats'))
@click.option('--secret', default=getDefaultSecretForApp('stats'))
@click.option('--raw', is_flag=True)
@click.option('--hide_nodes', is_flag=True)
@click.option('--hide_roles', is_flag=True)
@click.option('--subscribers', is_flag=True)
@click.option('--role_filter')
@click.option('--system', is_flag=True)
def monitor(url, role, secret, raw, role_filter, hide_nodes, hide_roles,
            subscribers, system):
    '''Monitor cobra
    '''

    preventRootUsage()
    uvloop.install()

    credentials = createCredentials(role, secret)
    runMonitor(url, credentials, raw, role_filter, not hide_nodes,
               not hide_roles, subscribers, system)



'''Write to the cobra key value store

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio

import click
import uvloop

# from cobras.client.kv_store import readClient
from cobras.client.connection import Connection
from cobras.client.credentials import (createCredentials, getDefaultRoleForApp,
                                       getDefaultSecretForApp)
from cobras.common.apps_config import PUBSUB_APPKEY
from cobras.common.superuser import preventRootUsage

DEFAULT_URL = f'ws://127.0.0.1:8765/v2?appkey={PUBSUB_APPKEY}'


@click.command()
@click.option('--url', default=DEFAULT_URL)
@click.option('--role', default=getDefaultRoleForApp('pubsub'))
@click.option('--secret', default=getDefaultSecretForApp('pubsub'))
@click.option('--channel', default='sms_republished_v1_neo')
@click.option('--data', default='{"foo": "bar"}')
def write(url, role, secret, channel, data):
    '''Write to the cobra key value store
    '''

    preventRootUsage()
    uvloop.install()

    credentials = createCredentials(role, secret)

    async def handler(url, credentials, channel, data):
        connection = Connection(url, credentials, verbose=True)
        await connection.connect()
        await connection.write(channel, data)
        await connection.close()

    asyncio.get_event_loop().run_until_complete(handler(url, credentials, channel, data))




'''Handle credential files

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import os

from cobras.common.apps_config import AppsConfig, getDefaultAppsConfigPath


def createCredentials(role, secret):
    return {'role': role, 'secret': secret}


def getDefaultRoleForApp(app) -> str:
    path = os.getenv('COBRA_APPS_CONFIG', getDefaultAppsConfigPath())
    appsConfig = AppsConfig(path)
    return appsConfig.getDefaultRoleForApp(app)


def getDefaultSecretForApp(app) -> str:
    path = os.getenv('COBRA_APPS_CONFIG', getDefaultAppsConfigPath())
    appsConfig = AppsConfig(path)
    return appsConfig.getDefaultSecretForApp(app)



'''App configuration, persisted in a yaml file. Credentials + some config

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import base64
import logging
import os
import base64
import tempfile
import gzip
import binascii
import io
from pathlib import Path
from random import getrandbits, randint
from typing import List

import yaml

STATS_APPKEY = '_stats'
HEALTH_APPKEY = '_health'
ADMIN_APPKEY = '_admin'
PUBSUB_APPKEY = '_pubsub'


class AppsConfig():
    def __init__(self, path: str) -> None:
        self.path = path
        self.data = {}
        self.apps = {}

        if os.path.exists(self.path):
            with open(self.path) as f:
                self.data = yaml.load(f.read(), Loader=yaml.FullLoader) or {}
                self.apps = self.data.get('apps')

            self.validateConfig()
        else:
            msg = f'Apps config file does not exists: "{path}". '
            msg += 'Use `cobra init` to create a default config file'
            logging.warning(msg)

    def validateConfig(self):
        if self.apps is None:
            raise ValueError(f'No apps present in config file')

        for app in self.apps:
            for roleName, role in self.apps[app]['roles'].items():
                if not isinstance(role, dict):
                    raise ValueError(f'role "{roleName}" is not a dict')

                if role.get('secret') is None:
                    raise ValueError(f'role "{roleName}" is missing a secret')

    def isAppKeyValid(self, appkey: str) -> bool:
        return self.apps.get(appkey) is not None

    def getRoleSecret(self, appkey: str, role: str) -> bytes:
        if not self.isAppKeyValid(appkey):
            logging.warning(f'Missing appkey: "{appkey}"')
            raise KeyError

        roles = self.apps.get(appkey, {}).get('roles')
        if roles is None:
            logging.warning(f'Missing roles')
            raise KeyError

        secret = roles.get(role, {}).get('secret')
        if secret is None:
            logging.warning(f'Missing role: "{role}"')
            raise KeyError

        return secret.encode('ascii')

    def getPermissions(self, appkey: str, role: str) -> List[str]:
        if not self.isAppKeyValid(appkey):
            logging.warning(f'Missing appkey: "{appkey}"')
            raise KeyError

        roles = self.apps.get(appkey, {}).get('roles')
        if roles is None:
            logging.warning(f'Missing roles')
            raise KeyError

        permissions = roles.get(role, {}).get('permissions', [])
        return permissions

    def isBatchPublishEnabled(self, appkey: str) -> bytes:
        if not self.isAppKeyValid(appkey):
            logging.warning(f'Missing appkey: "{appkey}"')
            raise KeyError

        batchPublish = self.apps.get(appkey, {}).get('batch_publish', False)
        return batchPublish

    def getBatchPublishSize(self):
        return self.data.get('batch_publish_size', -1)

    def generateDefaultConfig(self):
        self.data['apps'] = {}

        for app in [STATS_APPKEY, HEALTH_APPKEY, ADMIN_APPKEY, PUBSUB_APPKEY]:

            # Create an app with a user/role
            self.data['apps'][app] = {}
            self.data['apps'][app]['roles'] = {}

            role = app[1:]  # remove the leading _

            self.data['apps'][app]['roles'][role] = {
                'secret': genSecret(),
                'permissions': ['subscribe', 'publish', 'admin', 'read', 'write']
            }

        # write to disk
        with open(self.path, 'w') as f:
            yaml.dump(self.data, f, default_flow_style=False)

        # dump to the console (we could pretty-print)
        with open(self.path) as f:
            print(f.read())

    def getDefaultRoleForApp(self, app) -> str:
        try:
            roles = self.data['apps'][f'_{app}'].get('roles', {})
        except KeyError:
            return ''

        try:
            roleName = list(roles.keys())[0]
        except IndexError:
            return ''

        return roleName

    def getDefaultSecretForApp(self, app) -> str:
        try:
            roles = self.data['apps'][f'_{app}'].get('roles', {})
        except KeyError:
            return ''

        try:
            roleName = list(roles.keys())[0]
        except IndexError:
            return ''

        role = roles.get(roleName, {})
        return role.get('secret', '')


def generateAppsConfig(apps_config_path_content) -> str:
    tempPath = tempfile.mktemp(suffix='_apps.yaml')
    try:
        content = base64.b64decode(apps_config_path_content)
    except binascii.Error:
        return ''

    with open(tempPath, 'wb') as f:
        try:
            decompressed = gzip.GzipFile(fileobj=io.BytesIO(content)).read()
        except OSError:
            return ''

        f.write(decompressed)

    return tempPath


def getDefaultAppsConfigPath():
    content = os.getenv('COBRA_APPS_CONFIG_CONTENT')

    if content:
        apps_config_path = generateAppsConfig(content)
        os.environ['COBRA_APPS_CONFIG'] = apps_config_path
        return apps_config_path

    path = Path.home() / '.cobra.yaml'
    return str(path)


def generateNonce():
    bits = getrandbits(64)
    return base64.b64encode(bytearray(str(bits), 'utf8')).decode('ascii')


def genSecret():
    lowerChars = 'abcdef'
    upperChars = 'ABCDEF'
    nums = '0123456789'

    chars = lowerChars + upperChars + nums
    N = len(chars)

    secret = ''
    for _ in range(32):
        c = randint(0, N - 1)
        secret += chars[c]

    return secret



'''Client for a cobra server.

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio
import functools
import itertools
import json

import click
import websockets

from cobras.client.connection import Connection, AuthException, HandshakeException, ActionException


async def client(url, creds, clientCallback):
    '''Main client. Does authenticate then invoke the clientCallback which
    takes control.
    '''

    while True:
        try:
            connection = Connection(url, creds)
            await connection.connect()
            return await clientCallback(connection)

        except TimeoutError as e:
            click.secho(str(e), fg='red')
            await asyncio.sleep(1)
            pass
        except ConnectionRefusedError as e:
            click.secho(str(e), fg='red')
            await asyncio.sleep(1)
            pass
        except ConnectionResetError as e:
            click.secho(str(e), fg='red')
            await asyncio.sleep(1)
            pass
        except websockets.exceptions.ConnectionClosed as e:
            click.secho(str(e), fg='red')
            await asyncio.sleep(1)
            pass
        except OSError as e:
            click.secho(str(e), fg='red')
            await asyncio.sleep(1)
            pass
        except AuthException as e:
            click.secho(str(e), fg='red')
            await asyncio.sleep(1)
            pass
        except ActionException as e:
            click.secho(str(e), fg='red')
            await asyncio.sleep(1)
            pass


async def subscribeHandler(connection, **args):
    channel = args['channel']
    position = args['position']
    fsqlFilter = args['fsqlFilter']
    messageHandlerClass = args['messageHandlerClass']
    messageHandlerArgs = args['messageHandlerArgs']
    subscriptionId = args.get('subscription_id', channel)
    messageHandlerArgs['subscription_id'] = subscriptionId

    return await connection.subscribe(channel,
                                      position,
                                      fsqlFilter,
                                      messageHandlerClass,
                                      messageHandlerArgs,
                                      subscriptionId)


async def subscribeClient(url, credentials, channel, position, fsqlFilter,
                          messageHandlerClass, messageHandlerArgs):
    subscribeHandlerPartial = functools.partial(
        subscribeHandler,
        channel=channel,
        position=position,
        fsqlFilter=fsqlFilter,
        messageHandlerClass=messageHandlerClass,
        messageHandlerArgs=messageHandlerArgs)

    ret = await client(url, credentials, subscribeHandlerPartial)
    return ret


async def unsafeSubcribeClient(url, credentials, channel, position, fsqlFilter,
                               messageHandlerClass, messageHandlerArgs):
    '''
    No retry or exception handling
    Used by the health check, where we want to die hard and fast if there's a problem
    '''
    connection = Connection(url, credentials)
    await connection.connect()
    message = await connection.subscribe(channel,
                                         position,
                                         fsqlFilter,
                                         messageHandlerClass,
                                         messageHandlerArgs,
                                         subscriptionId=channel)
    return message


async def readHandler(websocket, **args):
    position = args.get('position')
    channel = args.get('channel')
    handler = args.get('handler')

    readPdu = {
        "action": "rtm/read",
        "body": {
            "channel": channel,
        },
        "id": 3  # FIXME
    }

    if position is not None:
        readPdu['body']['position'] = position

    print(f"> {readPdu}")
    await websocket.send(json.dumps(readPdu))

    readResponse = await websocket.recv()
    print(f"< {readResponse}")

    data = json.loads(readResponse)
    msg = data['body']['message']  # FIXME data missing
    await handler(msg)


async def readClient(url, credentials, channel, position, handler):
    readHandlerPartial = functools.partial(readHandler,
                                           channel=channel,
                                           position=position,
                                           handler=handler)

    ret = await client(url, credentials, readHandlerPartial)
    return ret



'''Compute our authentication hash with an hmac

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import base64
import hashlib
import hmac


def computeHash(secret: bytes, nonce: bytes) -> str:
    binary_hash = hmac.new(secret, nonce, hashlib.md5).digest()
    ascii_hash = base64.b64encode(binary_hash)

    h = ascii_hash.decode('ascii')

    # print(f'hmac({secret}, {nonce}) => {h}')
    return h



'''djb2 hash (unsused)

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''


def djb2Hash(data):
    '''Return a 32 bits digest'''

    hash = 5381
    for x in data:
        hash = ((hash << 5) + hash) + ord(x)
    return hash & 0xFFFFFFFF



'''Capture memory usage

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import os

import psutil

from cobras.common.memoize import memoize


@memoize
def getContainerMemoryLimit():
    '''In bytes'''

    path = '/sys/fs/cgroup/memory/memory.limit_in_bytes'
    if not os.path.exists(path):
        return psutil.virtual_memory().total
    else:
        with open(path) as f:
            mem = f.read()
            return int(mem)


def getProcessUsedMemory():
    p = psutil.Process()
    return p.memory_info().rss



'''Run the cobra frontend

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import os
import sys
import logging

import click
import sentry_sdk
import uvloop

from cobras.common.apps_config import getDefaultAppsConfigPath, generateAppsConfig
from cobras.common.superuser import preventRootUsage
from cobras.server.app import AppRunner


@click.command()
@click.option('--host',
              envvar='COBRA_HOST',
              default='127.0.0.1',
              help='Binding host address. Set to 0.0.0.0 in prod environments')
@click.option('--port', envvar='COBRA_PORT', default='8765')
@click.option('--redis_urls',
              envvar='COBRA_REDIS_URLS',
              default='redis://localhost;redis://localhost')
@click.option('--redis_password', envvar='COBRA_REDIS_PASSWORD')
@click.option('--apps_config_path',
              envvar='COBRA_APPS_CONFIG',
              default=getDefaultAppsConfigPath())
@click.option('--apps_config_path_content', envvar='COBRA_APPS_CONFIG_CONTENT')
@click.option('--prod', envvar='COBRA_PROD', is_flag=True)
@click.option('--plugins', envvar='COBRA_PLUGINS')
@click.option('--debug_memory', envvar='COBRA_DEBUG_MEMORY', is_flag=True)
@click.option('--sentry', envvar='COBRA_SENTRY', is_flag=True)
@click.option('--sentry_url', envvar='COBRA_SENTRY_URL')
@click.option('--no_stats', envvar='COBRA_NO_STATS', is_flag=True)
@click.option('--max_subscriptions',
              envvar='COBRA_MAX_SUSBSCRIPTIONS',
              default=-1)
@click.option('--idle_timeout',
              envvar='COBRA_IDLE_TIMEOUT',
              default=5 * 60,
              help='idle connections kicked out after X seconds')
def run(host, port, redis_urls, redis_password, apps_config_path,
        apps_config_path_content, debug_memory, plugins, sentry,
        sentry_url, prod, no_stats, max_subscriptions, idle_timeout):
    '''Run the cobra server

    \b
    cobra run --redis_urls 'redis://localhost:7001;redis://localhost:7002'
    \b
    env COBRA_REDIS_PASSWORD=foobared cobra run
    '''
    if prod:
        os.environ['COBRA_PROD'] = '1'

    preventRootUsage()
    uvloop.install()

    if sentry and sentry_url:
        sentry_sdk.init(sentry_url)

    if apps_config_path_content:
        apps_config_path = generateAppsConfig(apps_config_path_content)
        if not apps_config_path:
            logging.error(f'Invalid apps config path content: {apps_config_path_content}')
            logging.error(f'(usually configured with ${COBRA_APPS_CONFIG_CONTENT})')
            logging.error(f'Generate it with `gzip -c ~/.cobra.yaml | base64`')
            sys.exit(1)

        apps_config_path_content = '<cleared>'
        os.environ['COBRA_APPS_CONFIG'] = apps_config_path

    print('runServer', locals())
    runner = AppRunner(host, port, redis_urls, redis_password,
                       apps_config_path, debug_memory, plugins,
                       not no_stats, max_subscriptions, idle_timeout)
    try:
        runner.run()
    except OSError as e:
        logging.fatal(f'Cannot start cobra server: {e}')
        sys.exit(1)



'''Class to used to throttle function calls, so that something isn't called too
often

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

from datetime import datetime, timedelta


class Throttle(object):
    def __init__(self, seconds: int = 1, minutes: int = 0,
                 hours: int = 0) -> None:
        self.throttle_period = timedelta(seconds=seconds,
                                         minutes=minutes,
                                         hours=hours)
        self.time_of_last_call = datetime.min

    def exceedRate(self) -> bool:
        now = datetime.now()
        time_since_last_call = now - self.time_of_last_call

        if time_since_last_call > self.throttle_period:
            self.time_of_last_call = now
            return False
        else:
            return True



'''Publishers management

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio

from cobras.server.pipelined_publisher import PipelinedPublisher
from cobras.server.redis_connections import RedisConnections


class PipelinedPublishers():
    def __init__(self, redisConnections: RedisConnections,
                 batchPublishSize: int) -> None:
        self.redisConnections = redisConnections
        self.pipelinedPublishers: dict = {}
        self.batchPublishSize: int = batchPublishSize
        self.lock = asyncio.Lock()

    async def get(self, appkey, channel):
        '''
        Constraints:
        * (A) For a given app, subscriptions and publish should go to the
          same redis node.
        * (B) Each app should get its own redis connection,
          so that one 'busy' app does not block another one.

        Those constraints are reflected in the way we compute the hashing key
        '''
        # Constraint A
        appChannel = '{}::{}'.format(appkey, channel)

        # Constraint B
        key = '{}::{}'.format(appkey,
                              self.redisConnections.hashChannel(appChannel))

        async with self.lock:
            pipelinedPublisher = self.pipelinedPublishers.get(key)
            if pipelinedPublisher is not None:
                return pipelinedPublisher

            db = await self.redisConnections.create(appChannel)
            pipelinedPublisher = PipelinedPublisher(db, self.batchPublishSize)
            self.pipelinedPublishers[key] = pipelinedPublisher
            return pipelinedPublisher



'''Bavarde main driver.
   Calls into sub commands like git.

Copyright (c) 2019 Machine Zone, Inc. All rights reserved.
'''

from pkgutil import walk_packages

import click
import coloredlogs

LOGGING_FORMAT = '%(asctime)s %(levelname)s %(message)s'
coloredlogs.install(level='WARNING', fmt=LOGGING_FORMAT)


@click.option('--verbose', '-v', envvar='BAVARDE_VERBOSE', is_flag=True)
@click.group()
@click.version_option()
def main(verbose):
    """\b

Bavarde is a chat app written on top of cobra
    """
    if verbose:
        coloredlogs.install(level='INFO', fmt=LOGGING_FORMAT)


for loader, module_name, is_pkg in walk_packages(__path__, __name__ + '.'):
    module = __import__(module_name, globals(), locals(), ['__name__'])
    cmd = getattr(module, module_name.rsplit('.', 1)[-1])
    if isinstance(cmd, click.Command):
        main.add_command(cmd)



'''Capture statistics (publisher, subscribers, sent+received bytes, message
count, etc...). Used by `cobra monitor`

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio
import collections
import datetime
import json
import os
import platform
import time
import logging

from cobras.common.memory_usage import (getContainerMemoryLimit,
                                        getProcessUsedMemory)

DEFAULT_STATS_CHANNEL = '/stats'


class ServerStats():
    def __init__(self, pipelinedPublishers, appkey):
        self.pipelinedPublishers = pipelinedPublishers

        self.node = platform.uname().node
        self.connectionCount = 0
        self.stop = False

        self.idleConnections = 0

        self.internalAppKey = appkey
        self.statsChannel = DEFAULT_STATS_CHANNEL

        self.publishedCount = collections.defaultdict(int)
        self.publishedBytes = collections.defaultdict(int)
        self.subscribedCount = collections.defaultdict(int)
        self.subscribedBytes = collections.defaultdict(int)
        self.subscriptions = collections.defaultdict(int)

        self.readsCount = collections.defaultdict(int)
        self.readsBytes = collections.defaultdict(int)
        self.writesCount = collections.defaultdict(int)
        self.writesBytes = collections.defaultdict(int)

        self.resetCounterByPeriod()
        self.start = time.time()

        print('node ' + self.node)

    def incrConnections(self, role):
        self.connectionCount += 1

    def decrConnections(self, role):
        self.connectionCount -= 1

    def incrIdleConnections(self):
        self.idleConnections += 1

    def incrSubscriptions(self, role):
        self.subscriptions[role] += 1

    def decrSubscriptionsBy(self, role, subscriptionsCount):
        self.subscriptions[role] -= subscriptionsCount

    def resetCounterByPeriod(self):
        self.publishedCountByPeriod = collections.defaultdict(int)
        self.publishedBytesByPeriod = collections.defaultdict(int)
        self.subscribedCountByPeriod = collections.defaultdict(int)
        self.subscribedBytesByPeriod = collections.defaultdict(int)

        self.readsCountByPeriod = collections.defaultdict(int)
        self.readsBytesByPeriod = collections.defaultdict(int)
        self.writesCountByPeriod = collections.defaultdict(int)
        self.writesBytesByPeriod = collections.defaultdict(int)

    def updatePublished(self, role, val):
        self.publishedCount[role] += 1
        self.publishedBytes[role] += val

        self.publishedCountByPeriod[role] += 1
        self.publishedBytesByPeriod[role] += val

    def updateSubscribed(self, role, val):
        self.subscribedCount[role] += 1
        self.subscribedBytes[role] += val

        self.subscribedCountByPeriod[role] += 1
        self.subscribedBytesByPeriod[role] += val

    def updateReads(self, role, val):
        self.readsCount[role] += 1
        self.readsBytes[role] += val

        self.readsCountByPeriod[role] += 1
        self.readsBytesByPeriod[role] += val

    def updateWrites(self, role, val):
        self.writesCount[role] += 1
        self.writesBytes[role] += val

        self.writesCountByPeriod[role] += 1
        self.writesBytesByPeriod[role] += val

    async def run(self):
        while True:
            # Only dict-like objects are permitted in that field
            # to ease the job of aggregating them in the monitor command
            cobraData = {
                'subscriptions': self.subscriptions,
            }

            cobraData.update({
                'published_count': self.publishedCount,
                'published_bytes': self.publishedBytes,
                'published_count_per_second': self.publishedCountByPeriod,
                'published_bytes_per_second': self.publishedBytesByPeriod
            })

            cobraData.update({
                'subscribed_count': self.subscribedCount,
                'subscribed_bytes': self.subscribedBytes,
                'subscribed_count_per_second': self.subscribedCountByPeriod,
                'subscribed_bytes_per_second': self.subscribedBytesByPeriod,
            })

            cobraData.update({
                'reads_count': self.readsCount,
                'reads_bytes': self.readsBytes,
                'reads_count_per_second': self.readsCountByPeriod,
                'reads_bytes_per_second': self.readsBytesByPeriod
            })

            cobraData.update({
                'writes_count': self.writesCount,
                'writes_bytes': self.writesBytes,
                'writes_count_per_second': self.writesCountByPeriod,
                'writes_bytes_per_second': self.writesBytesByPeriod
            })

            uptime = time.time() - self.start
            uptimeMinutes = uptime // 60
            uptime = str(datetime.timedelta(seconds=uptime))
            uptime, _, _ = uptime.partition('.')  # skip the milliseconds part

            message = {
                'node': self.node,
                'prod': os.getenv('COBRA_PROD') is not None,
                'data': {
                    'cobra': cobraData,
                    'system': {
                        'connections': self.connectionCount,
                        'mem_bytes': getProcessUsedMemory(),
                        'container_memory_limit_bytes':
                        getContainerMemoryLimit(),  # noqa
                        'uptime': uptime,
                        'uptime_minutes': uptimeMinutes,
                        'tasks': len(asyncio.all_tasks()),
                        'idle_connections': self.idleConnections
                    }
                }
            }

            data = json.dumps({'body': {'message': message}})

            chan = self.statsChannel
            appkey = self.internalAppKey

            try:
                pipelinedPublisher = await self.pipelinedPublishers.get(appkey, chan)
                await pipelinedPublisher.publishNow((appkey, chan, data))
            except Exception as e:
                logging.error(f"stats: cannot connect to redis {e}")
                pass

            self.resetCounterByPeriod()

            # Sleep 1 second
            await asyncio.sleep(1)

            if self.stop:
                return

    def terminate(self):
        self.stop = True



'''Health check to validate that a server is ok.

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio
import json
import uuid
import os
from typing import Dict

from cobras.client.client import subscribeClient, unsafeSubcribeClient
from cobras.client.connection import Connection
from cobras.client.credentials import createCredentials
from cobras.common.apps_config import HEALTH_APPKEY


def getDefaultHealthCheckChannel():
    randomKey = uuid.uuid4().hex[:8]
    return f'sms_health_check_channel_{randomKey}'


def getDefaultHealthCheckHttpUrl(host=None, port=None):
    if host is None:
        host = '127.0.0.1'
    if port is None:
        port = os.getenv('COBRA_PORT', 8765)

    return f'http://{host}:{port}/health/'


def getDefaultHealthCheckUrl(host=None, port=None):
    if host is None:
        host = '127.0.0.1'
    if port is None:
        port = os.getenv('COBRA_PORT', 8765)

    return f'ws://{host}:{port}/v2?appkey={HEALTH_APPKEY}'


def healthCheck(url, role, secret, channel, retry=False):
    '''Perform a health check'''
    class MessageHandlerClass:
        def __init__(self, connection, args):
            self.connection = connection
            self.channel = args['channel']
            self.content = args['content']
            self.subscriptionId = self.channel
            self.parsedMessage = None

        async def on_init(self):
            await self.connection.publish(self.channel, self.content)

        async def handleMsg(self, message: Dict, position: str) -> bool:
            self.parsedMessage = message
            return False

    credentials = createCredentials(role, secret)

    position = None
    magicNumber = 666
    refAndroidId = uuid.uuid4().hex
    content = {
        'device': {
            'game': 'test',
            'android_id': refAndroidId
        },
        'magic': magicNumber
    }

    fsqlFilter = f"""select magic,device.android_id
                     from `{channel}` where
                         device.game = 'test' AND
                         device.android_id = '{refAndroidId}' AND
                         magic = {magicNumber}
    """

    messageHandlerArgs = {
        'channel': channel,
        'content': content
    }

    handler = unsafeSubcribeClient
    if retry:
        handler = subscribeClient

    messageHandler = asyncio.get_event_loop().run_until_complete(
        handler(url, credentials, channel, position, fsqlFilter,
                MessageHandlerClass, messageHandlerArgs))

    message = messageHandler.parsedMessage

    if message.get('device.android_id') != refAndroidId:
        raise ValueError(f'incorrect android_id: {data}')

    if magicNumber != message['magic']:
        raise ValueError(f'incorrect magic: {data}')



'''Subscribe to a channel

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import logging
import asyncio
import json
from typing import Dict

import click
import uvloop

from cobras.client.client import subscribeClient
from cobras.client.credentials import (createCredentials, getDefaultRoleForApp,
                                       getDefaultSecretForApp)
from cobras.common.apps_config import PUBSUB_APPKEY
from cobras.common.superuser import preventRootUsage
from cobras.common.throttle import Throttle

DEFAULT_URL = f'ws://127.0.0.1:8765/v2?appkey={PUBSUB_APPKEY}'


class MessageHandlerClass:
    def __init__(self, websockets, args):
        self.cnt = 0
        self.cntPerSec = 0
        self.throttle = Throttle(seconds=1)

    async def on_init(self):
        pass

    async def handleMsg(self, message: Dict, position: str) -> bool:
        self.cnt += 1
        self.cntPerSec += 1

        logging.info(f'{message} at position {position}')

        if self.throttle.exceedRate():
            return True

        print(f"#messages {self.cnt} msg/s {self.cntPerSec}")
        self.cntPerSec = 0

        return True


@click.command()
@click.option('--url', default=DEFAULT_URL)
@click.option('--role', default=getDefaultRoleForApp('pubsub'))
@click.option('--secret', default=getDefaultSecretForApp('pubsub'))
@click.option('--channel', default='sms_republished_v1_neo')
@click.option('--position')
@click.option('--stream_sql')
def subscribe(url, role, secret, channel, position, stream_sql):
    '''Subscribe to a channel
    '''

    preventRootUsage()
    uvloop.install()

    credentials = createCredentials(role, secret)

    asyncio.get_event_loop().run_until_complete(
        subscribeClient(url, credentials, channel, position, stream_sql,
                        MessageHandlerClass, {}))



'''Cobra connection

Copyright (c) 2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio
import itertools
import json
import collections
import logging

import websockets

from cobras.common.auth_hash import computeHash


class AuthException(Exception):
    pass


class HandshakeException(Exception):
    pass


class ActionException(Exception):
    pass


class Connection(object):
    def __init__(self, url, creds):
        self.url = url
        self.creds = creds
        self.idIterator = itertools.count()

        # different queues per action kind or instances
        self.queues = collections.defaultdict(asyncio.Queue)

        self.task = None

    async def connect(self):
        self.websocket = await websockets.connect(self.url)
        self.task = asyncio.create_task(self.waitForResponses())
        self.stop = asyncio.get_running_loop().create_future()

        role = self.creds['role']

        handshake = {
            "action": "auth/handshake",
            "id": next(self.idIterator),
            "body": {
                "data": {
                    "role": self.creds['role']
                },
                "method": "role_secret"
            }
        }

        response = await self.send(handshake)
        nonce = bytearray(response['body']['data']['nonce'], 'utf8')
        secret = bytearray(self.creds['secret'], 'utf8')

        challenge = {
            "action": "auth/authenticate",
            "id": next(self.idIterator),
            "body": {
                "method": "role_secret",
                "credentials": {
                    "hash": computeHash(secret, nonce)
                }
            }
        }
        await self.send(challenge)

    def __del__(self):
        if self.task is not None:
            self.task.cancel()

    async def waitForResponses(self):
        try:
            while True:
                response = await self.websocket.recv()

                logging.debug(f'< {response}')
                data = json.loads(response)

                action = data['action']
                action = '/'.join(action.split('/')[:2])
                actionId = action + '::' + str(data['id'])

                if action == 'rtm/subscription':
                    actionId = action + '::' + data['body']['subscription_id']

                q = self.getQueue(actionId)

                await q.put(data)

        except Exception as e:
            logging.error(f'unexpected exception: {e}')

            # propagate the exception to the caller
            self.stop.set_result(e)

    def getQueue(self, action):
        q = self.queues[action]
        return q

    def computeDefaultActionId(self, pdu):
        action = pdu['action']
        actionId = f'{action}::' + str(pdu['id'])
        return actionId

    async def getActionResponse(self, actionId):
        '''
        This could be as simple as:
            data = await self.getQueue(actionId).get()

        Unfortunately we need to handle exception in the 
        'fetch response coroutine' so that we can rethrow them here
        '''
        incoming = asyncio.ensure_future(self.getQueue(actionId).get())

        done, pending = await asyncio.wait(
             [incoming, self.stop],
             return_when=asyncio.FIRST_COMPLETED)

        # Cancel pending tasks to avoid leaking them.
        if incoming in pending:
            incoming.cancel()

        if incoming in done:
            data = incoming.result()
            return data

        if self.stop in done:
            # Reraise the exception which was caught in self.waitForResponses
            raise self.stop.result()

    async def send(self, pdu):
        actionId = self.computeDefaultActionId(pdu)

        data = json.dumps(pdu)
        logging.info(f"> {data}")
        await self.websocket.send(data)

        # get the response
        data = await self.getActionResponse(actionId)
        logging.info(f"< {data}")

        # validate response
        if data.get('action') != (pdu['action'] + '/ok'):
            raise ActionException(data.get('body', {}).get('error'))

        return data

    async def subscribe(self,
                        channel,
                        position,
                        fsqlFilter,
                        messageHandlerClass,
                        messageHandlerArgs,
                        subscriptionId):
        pdu = {
            "action": "rtm/subscribe",
            "id": next(self.idIterator),
            "body": {
                "subscription_id": subscriptionId,
                "channel": channel,
                "fast_forward": True,
                "filter": fsqlFilter
            },
        }
        await self.send(pdu)

        messageHandler = messageHandlerClass(self, messageHandlerArgs)
        await messageHandler.on_init()

        actionId = 'rtm/subscription::' + subscriptionId

        while True:
            data = await self.getActionResponse(actionId)

            message = data['body']['messages'][0]
            position = data['body']['position']

            ret = await messageHandler.handleMsg(message, position)
            if not ret:
                break

        try:
            await self.unsubscribe(subscriptionId)
        except websockets.exceptions.ConnectionClosed as e:
            logging.warning(f"Connection is closed, cannot unsubscribe")
            pass

        return messageHandler

    async def unsubscribe(self, subscriptionId):
        pdu = {
            "action": "rtm/unsubscribe",
            "id": next(self.idIterator),
            "body": {
                "subscription_id": subscriptionId
            }
        }
        await self.send(pdu)

    async def publish(self, channel, msg):
        pdu = {
            "action": "rtm/publish",
            "id": next(self.idIterator),
            "body": {
                "channel": channel,
                "message": msg
            }
        }
        await self.send(pdu)

    async def write(self, channel, msg):
        pdu = {
            "action": "rtm/write",
            "id": next(self.idIterator),
            "body": {
                "channel": channel,
                "message": msg
            }
        }
        await self.send(pdu)

    async def read(self, channel, position=None):
        pdu = {
            "action": "rtm/read",
            "id": next(self.idIterator),
            "body": {
                "channel": channel,
            }
        }
        data = await self.send(pdu)

        msg = data['body']['message']  # FIXME data missing / error handling ?
        return msg

    async def close(self):
        await self.websocket.close()
        close_status = websockets.exceptions.format_close(self.websocket.close_code,
                                                          self.websocket.close_reason)
        return close_status




'''Chat client

Copyright (c) 2019 Machine Zone, Inc. All rights reserved.

Lot of code borrowed from websockets cli.
'''

import asyncio
import getpass
import os
import pprint
import threading
import sys

import click
import uvloop

from cobras.bavarde.client.client import runClient

DEFAULT_LOCAL_URL = 'ws://127.0.0.1:8765/v2?appkey=_pubsub'
DEFAULT_URL = 'ws://chat.jeanserge.com/v2?appkey=_pubsub'

DEFAULT_ROLE = 'pubsub'
DEFAULT_SECRET = 'ccc02DE4Ed8CAB9aEfC8De3e13BfBE5E'


@click.command()
@click.option('--url', default=DEFAULT_URL)
@click.option('--role', envvar='BAVARDE_DEFAULT_ROLE', default=DEFAULT_ROLE)
@click.option('--secret',
              envvar='BAVARDE_DEFAULT_SECRET',
              default=DEFAULT_SECRET,
              required=True)
@click.option('--channel', envvar='BAVARDE_DEFAULT_CHANNEL', default='lobby')
@click.option('--position', default='0-0')
@click.option('--username', default=getpass.getuser())
@click.option('--verbose', '-v', count=True)
@click.option('--dev', '-v', count=True)
@click.option('--stream_sql')
def client(url, role, secret, channel, position, username, stream_sql, verbose,
           dev):
    '''WRITEME'''

    uvloop.install()

    if os.getenv('DEBUG') is not None or verbose:
        pprint.pprint(locals())

    try:
        import readline  # noqa
    except ImportError:  # Windows has no `readline` normally
        pass

    if dev:
        url = DEFAULT_LOCAL_URL

    # Create an event loop that will run in a background thread.
    loop = asyncio.new_event_loop()

    # Create a queue of user inputs. There's no need to limit its size.
    inputs: asyncio.Queue[str] = asyncio.Queue(loop=loop)

    # Create a stop condition when receiving SIGINT or SIGTERM.
    stop: asyncio.Future[None] = loop.create_future()

    # Schedule the task that will manage the connection.
    asyncio.ensure_future(runClient(url, role, secret, channel, position,
                                    stream_sql, verbose, username, loop,
                                    inputs, stop),
                          loop=loop)

    # Start the event loop in a background thread.
    thread = threading.Thread(target=loop.run_forever)
    thread.start()

    # Read from stdin in the main thread in order to receive signals.
    try:
        while True:
            message = input('> ')

            # Erase that line we just typed. FIXME: handle multi-lines inputs
            sys.stdout.write(
                # Move cursor up
                "\N{ESC}[A"
                # Move cursor to beginning of line
                "\N{CARRIAGE RETURN}"
                # Delete current line
                "\N{ESC}[K")

            # Since there's no size limit, put_nowait is identical to put.
            loop.call_soon_threadsafe(inputs.put_nowait, message)
    except (KeyboardInterrupt, EOFError):  # ^C, ^D
        loop.call_soon_threadsafe(stop.set_result, None)

    # Wait for the event loop to terminate.
    thread.join()



'''Utility for batch publishing.

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.

FIXME: Should be renamed
'''

import json


def computeEventTimeDeltas(events: str) -> list:

    lines = events.splitlines()
    N = len(lines)
    assert N > 0, 'event list should not be empty'

    events = [json.loads(line) for line in lines]

    deltas = []
    for i in range(1, N):
        previousTs = events[i - 1]['timestamp']
        currentTs = events[i]['timestamp']
        delta = currentTs - previousTs

        deltas.append((lines[i - 1], delta))

    deltas.append((lines[N - 1], 0))

    return deltas



'''Chat client

Copyright (c) 2019 Machine Zone, Inc. All rights reserved.

Lot of code borrowed from websockets cli.
'''

import asyncio
import datetime
import json
import os
import pprint
import signal
import sys
import threading
import uuid
import datetime
from typing import Any, Set

import click
import websockets
from websockets.exceptions import format_close

from cobras.client.client import subscribeClient
from cobras.client.credentials import (createCredentials, getDefaultRoleForApp,
                                       getDefaultSecretForApp)
from cobras.common.throttle import Throttle


def exit_from_event_loop_thread(loop: asyncio.AbstractEventLoop,
                                stop: "asyncio.Future[None]") -> None:
    loop.stop()
    if not stop.done():
        # When exiting the thread that runs the event loop, raise
        # KeyboardInterrupt in the main thread to exit the program.
        try:
            ctrl_c = signal.CTRL_C_EVENT  # Windows
        except AttributeError:
            ctrl_c = signal.SIGINT  # POSIX
        os.kill(os.getpid(), ctrl_c)


def print_during_input(string: str) -> None:
    sys.stdout.write(
        # Save cursor position
        "\N{ESC}7"
        # Add a new line
        "\N{LINE FEED}"
        # Move cursor up
        "\N{ESC}[A"
        # Insert blank line, scroll last line down
        "\N{ESC}[L"
        # Print string in the inserted blank line
        f"{string}\N{LINE FEED}"
        # Restore cursor position
        "\N{ESC}8"
        # Move cursor down
        "\N{ESC}[B")
    sys.stdout.flush()


def print_over_input(string: str) -> None:
    sys.stdout.write(
        # Move cursor to beginning of line
        "\N{CARRIAGE RETURN}"
        # Delete current line
        "\N{ESC}[K"
        # Print string
        f"{string}\N{LINE FEED}")
    sys.stdout.flush()


def colorize(name):
    colors = [
        'red',
        'green',
        'yellow',
        'blue',
        'magenta',
        'cyan'
    ]
    idx = hash(name) % len(colors)
    color = colors[idx]
    return click.style(name, fg=color)


class MessageHandlerClass:
    def __init__(self, connection, args):
        self.verbose = args['verbose']
        self.q = args['queue']

        args['connection'] = connection

    async def on_init(self):
        '''Get a connection to the DB'''
        print('Ready to receive messages')

    async def handleMsg(self, message: dict, position: str) -> bool:
        self.q.put_nowait((message, position))
        return True


async def runClient(url, role, secret, channel, position, stream_sql, verbose,
                    username, loop, inputs, stop):
    credentials = createCredentials(role, secret)

    q: asyncio.Queue[str] = asyncio.Queue(loop=loop)

    args = {'verbose': verbose, 'queue': q}

    task = asyncio.create_task(
        subscribeClient(url, credentials, channel, position, stream_sql,
                        MessageHandlerClass, args))

    try:
        while True:
            incoming: asyncio.Future[Any] = asyncio.ensure_future(q.get())
            outgoing: asyncio.Future[Any] = asyncio.ensure_future(inputs.get())
            done: Set[asyncio.Future[Any]]
            pending: Set[asyncio.Future[Any]]
            done, pending = await asyncio.wait(
                [incoming, outgoing, stop],
                return_when=asyncio.FIRST_COMPLETED)

            # Cancel pending tasks to avoid leaking them.
            if incoming in pending:
                incoming.cancel()
            if outgoing in pending:
                outgoing.cancel()

            if incoming in done:
                try:
                    (message, position) = incoming.result()
                except websockets.ConnectionClosed:
                    break
                else:
                    data = message.get('data', {})
                    user = data.get('user', 'unknown user')
                    text = data.get('text', '<invalid message>')
                    messageId = message.get('id')

                    # Use redis position to get a datetime
                    timestamp = position.split('-')[0]
                    dt = datetime.datetime.fromtimestamp(int(timestamp) / 1000)
                    dtFormatted = dt.strftime('[%H:%M:%S]')

                    maxUserNameLength = 12
                    padding = (maxUserNameLength - len(user)) * ' '

                    user = colorize(user)
                    print_during_input(f'{dtFormatted} {padding} {user}: {text}')

            if outgoing in done:
                text = outgoing.result()

                messageId = uuid.uuid4().hex  # FIXME needed ?

                message = {
                    'data': {
                        'user': username,
                        'text': text
                    },
                    'id': messageId
                }

                await args['connection'].publish(channel, message)

            if stop in done:
                break

    finally:
        connection = args.get('connection')
        if connection is not None:
            closeStatus = await args['connection'].close()
            print_over_input(f"Connection closed: {closeStatus}.")

        task.cancel()
        await task

        exit_from_event_loop_thread(loop, stop)



'''Publish to a channel

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio
import functools
import json

import click
import uvloop

from cobras.client.client import client
from cobras.client.credentials import (createCredentials, getDefaultRoleForApp,
                                       getDefaultSecretForApp)
from cobras.common.apps_config import ADMIN_APPKEY
from cobras.common.superuser import preventRootUsage

DEFAULT_URL = f'ws://127.0.0.1:8765/v2?appkey={ADMIN_APPKEY}'


async def clientCallback(websocket, **args):
    method = args['method']
    params = args.get('params')
    params = json.loads(params)

    pdu = {
        "action": "rpc/admin",
        "id": 1,
        "body": {
            "method": method,
            "params": params
        }
    }

    await websocket.send(json.dumps(pdu))

    response = await websocket.recv()
    print(response)


async def start(url, credentials, method, params):
    callback = functools.partial(clientCallback,
                                 method=method,
                                 params=params)

    task = asyncio.create_task(client(url, credentials, callback))
    await task


@click.command()
@click.option('--url', default=DEFAULT_URL)
@click.option('--role', default=getDefaultRoleForApp('admin'))
@click.option('--secret', default=getDefaultSecretForApp('admin'))
@click.option('--method', default='close_all')
@click.option('--params', default='{}')
def admin(url, role, secret, method, params):
    '''Admin

    \b
    cobra admin --method toggle_file_logging --params '{"connection_id": ".."}'
    '''

    preventRootUsage()
    uvloop.install()

    credentials = createCredentials(role, secret)

    asyncio.get_event_loop().run_until_complete(
        start(url, credentials, method, params))



'''Compute our version number

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import pkg_resources

from cobras.common.memoize import memoize


def getVersion():
    '''Get our own version using pkg_resources'''

    return pkg_resources.get_distribution("cobras").version



'''Key Value store operations (set, get, delete),
but that are using streams for storage

Copyright (c) 2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio
import traceback
from typing import Optional

import ujson

from cobras.server.redis_connections import RedisConnections


async def kvStoreRead(redisConnections: RedisConnections, pattern: str,
                      position: Optional[str], logger):
    # Create connection
    connection = await redisConnections.create(pattern)

    if position is None:
        # Get the last entry written to a stream
        end = '-'
        start = '+'
    else:
        start = position
        end = start

    try:
        results = await connection.xrevrange(pattern, start, end, 1)
        if len(results) == 0:
            return None

        result = results[0]
        position = result[0]
        msg = result[1]
        data = msg[b'json']

        msg = ujson.loads(data)
        return msg

    except asyncio.CancelledError:
        logger('Cancelling redis subscription')
        raise

    except Exception as e:
        logger(e)
        logger('Generic Exception caught in {}'.format(traceback.format_exc()))

    finally:
        # When finished, close the connection.
        connection.close()
